const std = @import("std");

const get_sock_info = @import("libhoneybee").socket.get_sock_info;
const StateFile = @import("libhoneybee").state.StateFile;
const CallStatus = @import("libhoneybee").state.CallStatus;
const EventType = @import("libhoneybee").state.EventType;

const call = @import("snikket_call.zig");
const config = @import("config.zig");
const socket = @import("socket_service.zig");
const StateChange = @import("state.zig").StateChange;
const HookHandler = @import("state.zig").HookHandler;
const CallbackFifo = @import("state.zig").CallbackFifo;
const SocketFifo = @import("state.zig").SocketFifo;
const service_message = @import("state.zig").service_message;

pub fn main() !void {
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    defer _ = gpa.deinit();
    const allocator = gpa.allocator();

    const conf = try config.get_config(allocator);
    defer conf.free();

    var state_file = StateFile{
        .runtime_dir = conf.env_variables.xdg_runtime_dir,
        .state_path = conf.env_variables.state_path,
        .allocator = allocator,
    };
    state_file.create_path();
    // make sure state file advertises uninitialized before launching socket
    try state_file.update_state_file(CallStatus.uninitialized, "");

    var hook_handler = HookHandler{
        .hook_path = conf.env_variables.xdg_config_home,
        .allocator = allocator,
    };

    var mutex = std.Thread.Mutex{};
    var cond = std.Thread.Condition{};

    const callback_fifo = std.fifo.LinearFifo(StateChange, .Dynamic).init(allocator);
    var callbacks = CallbackFifo{ .q = callback_fifo, .mutex = &mutex, .cond = &cond };
    defer callbacks.deinit();

    const socket_fifo = std.fifo.LinearFifo(socket.Event, .Dynamic).init(allocator);
    var sockets_messages = SocketFifo{ .q = socket_fifo, .mutex = &mutex, .cond = &cond };
    defer sockets_messages.deinit();

    const sock_const = get_sock_info();
    var sock = socket.SockHost{ .address = sock_const.address, .sock = sock_const.sock, .allocator = allocator, .messages = &sockets_messages };
    try sock.connect();
    _ = try std.Thread.spawn(.{}, socket_service, .{&sock});

    var snikket_call = call.CallState{ .messages = &callbacks };
    snikket_call.init(conf.json.value.username, conf.password);
    std.debug.assert(callbacks.len() == 1);
    try service_message(&callbacks, &hook_handler, &state_file, &snikket_call);
    std.debug.assert(callbacks.len() == 0);
    snikket_call.connect();
    std.debug.assert(callbacks.len() == 1);
    try service_message(&callbacks, &hook_handler, &state_file, &snikket_call);

    while (true) {
        mutex.lock();
        while (callbacks.len() == 0 and sockets_messages.len() == 0) {
            std.log.debug("waiting", .{});
            cond.wait(&mutex);
            std.log.debug("done waiting", .{});
        }
        std.log.debug("woke up!", .{});

        // don't hold onto mutex while processing job
        mutex.unlock();
        while (callbacks.len() > 0) {
            try service_message(&callbacks, &hook_handler, &state_file, &snikket_call);
        }
        while (sockets_messages.len() > 0) {
            var event = sockets_messages.next_message();
            defer event.free();

            switch (event.event_type) {
                EventType.call => snikket_call.call(event.payload),
                EventType.hangup => snikket_call.hangup(event.payload),
                EventType.accept_call => snikket_call.accept_call(event.payload),
                EventType.dtmf => snikket_call.send_dtmf(event.payload),
                else => std.log.info("Invalid event: {}\n", .{event}),
            }
        }
    }

    gpa.deinit();
}

fn socket_service(sock: *socket.SockHost) !void {
    while (true) {
        try sock.block_till_event();
    }
}
