/**
 * \file snikket.h the Snikket SDK for C
 *
 * Everything returned from an SDK procedure or passed to a function
 * pointer, both strings and opaque types, must be passed to
 * snikket_release when you are done with it.
 */

#ifndef __SNIKKET_H
#define __SNIKKET_H

#ifndef MKDOCS
#include <stdbool.h>
#include <stddef.h>

#ifndef API_PREFIX
	#ifdef _WIN32
		#define API_PREFIX __declspec(dllimport)
	#else
		#define API_PREFIX
	#endif
#endif

#endif

typedef void (*snikket_panic_callback) (const char *info);

enum
#ifdef __clang__
__attribute__((enum_extensibility(closed)))
#endif
snikket_ui_state {
	Pinned = 0,
	Open = 1,
	Closed = 2
};

enum
#ifdef __clang__
__attribute__((enum_extensibility(closed)))
#endif
snikket_message_direction {
	MessageReceived = 0,
	MessageSent = 1
};

enum
#ifdef __clang__
__attribute__((enum_extensibility(closed)))
#endif
snikket_message_status {
	MessagePending = 0,
	MessageDeliveredToServer = 1,
	MessageDeliveredToDevice = 2,
	MessageFailedToSend = 3
};

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Initializes the SDK
 *
 * This must be first before calling SDK functions (otherwise those calls will hang waiting for a response).
 *
 * @param panicCallback a callback to execute if the SDK panics. The SDK will continue processing events after a panic and you may want to stop it after receiving this callback. Use `NULL` for no callback
 * @returns `NULL` if the thread initializes successfully or a null-terminated C string if an error occurs during initialization
 */
API_PREFIX const char *snikket_setup(snikket_panic_callback panic_callback);

/**
 * Stops the SDK, blocking until the main thread has completed. Once ended, it cannot be restarted (this is because static variable state will be retained from the last run).
 *
 * It can be safely called any number of times – if the SDK is not running this function will just return.
 *
 * After executing no more calls to SDK functions can be made (as these will hang waiting for a response).
 *
 * Thread-safety: Can be called safely called on any thread.
 *
 * @param wait If `true`, this function will wait for all events scheduled to execute in the future on the SDK thread to complete. If `false`, immediate pending events will be finished and the SDK stopped without executing events scheduled in the future
 */
API_PREFIX void snikket_stop(bool wait);

/**
 * Informs the garbage collector that object is no longer needed by the C code.
 * 
 * If the object has no remaining reference the garbage collector can free the associated memory (which can happen at any time in the future). It does not free the memory immediately.
 * 
 * Thread-safety: can be called on any thread.
 * 
 * @param ptr a handle to an arbitrary SDK object returned from an SDK function
 */
API_PREFIX void snikket_release(const void *ptr);

/**
 * Create a basic persistence layer based on sqlite
 * 
 * @param dbfile path to sqlite database
 * @params blobpath path to directory for blob storage
 * @returns new persistence layer
 */
API_PREFIX void *snikket_persistence_sqlite_new(const char *dbfile, const char *blobpath);

/**
 * Create a basic persistence layer that persists nothing
 * 
 * @returns new persistence layer
 */
API_PREFIX void *snikket_persistence_dummy_new();

/**
 * Create a new Client to connect to a particular account
 * 
 * @param address the account to connect to
 * @param persistence the persistence layer to use for storage
 */
API_PREFIX void *snikket_client_new(const char *address, void *persistence);

/**
 * Set to false to suppress sending available presence
 */
API_PREFIX void snikket_client_set_send_available(void *client, bool value);

/**
 * Start this client running and trying to connect to the server
 */
API_PREFIX void snikket_client_start(void *client);

/**
 * Sets the password to be used in response to the password needed event
 * 
 * @param password
 */
API_PREFIX void snikket_client_use_password(void *client, const char *password);

/**
 * Get the account ID for this Client
 * 
 * @returns account id
 */
API_PREFIX const char *snikket_client_account_id(void *client);

/**
 * Get the current display name for this account
 * 
 * @returns display name
 */
API_PREFIX const char *snikket_client_display_name(void *client);

/**
 * Set the current display name for this account on the server
 * 
 * @param display name to set (ignored if empty or NULL)
 */
API_PREFIX void snikket_client_set_display_name(void *client, const char *displayName);

/**
 * @returns array of open chats, sorted by last activity
 */
API_PREFIX size_t snikket_client_get_chats(void *client, void ***outPtr);

/**
 * Search for chats the user can start or join
 * 
 * @param q the search query to use
 * @param callback takes two arguments, the query that was used and the array of results
 */
API_PREFIX void snikket_client_find_available_chats(void *client, const char *q, void (*callback) (const char*, void**, size_t, void*), void *callback__context);

/**
 * Start or join a chat from the search results
 * 
 * @returns the chat that was started
 */
API_PREFIX void *snikket_client_start_chat(void *client, void *availableChat);

/**
 * Find a chat by id
 * 
 * @returns the chat if known, or NULL
 */
API_PREFIX void *snikket_client_get_chat(void *client, const char *chatId);

/**
 * Event fired when client needs a password for authentication
 * 
 * @param handler takes one argument, the Client that needs a password
 */
API_PREFIX void snikket_client_add_password_needed_listener(void *client, void (*handler) (void*, void*), void *handler__context);

/**
 * Event fired when client is connected and fully synchronized
 * 
 * @param handler takes no arguments
 */
API_PREFIX void snikket_client_add_status_online_listener(void *client, void (*handler) (void*), void *handler__context);

/**
 * Event fired when client is disconnected
 * 
 * @param handler takes no arguments
 */
API_PREFIX void snikket_client_add_status_offline_listener(void *client, void (*handler) (void*), void *handler__context);

/**
 * Event fired when a new ChatMessage comes in on any Chat
 * Also fires when status of a ChatMessage changes,
 * when a ChatMessage is edited, or when a reaction is added
 * 
 * @param handler takes one argument, the ChatMessage
 */
API_PREFIX void snikket_client_add_chat_message_listener(void *client, void (*handler) (void*, void*), void *handler__context);

/**
 * Event fired when a Chat's metadata is updated, or when a new Chat is added
 * 
 * @param handler takes one argument, an array of Chats that were updated
 */
API_PREFIX void snikket_client_add_chats_updated_listener(void *client, void (*handler) (void**, size_t, void*), void *handler__context);

/**
 * Event fired when a new call comes in
 * 
 * @param handler takes two arguments, the call Session and the associated Chat ID
 */
API_PREFIX void snikket_client_add_call_ring_listener(void *client, void (*handler) (void*, const char*, void*), void *handler__context);

/**
 * Event fired when a call is retracted or hung up
 * 
 * @param handler takes one argument, the associated Chat ID
 */
API_PREFIX void snikket_client_add_call_retract_listener(void *client, void (*handler) (const char*, void*), void *handler__context);

/**
 * Event fired when an outgoing call starts ringing
 * 
 * @param handler takes one argument, the associated Chat ID
 */
API_PREFIX void snikket_client_add_call_ringing_listener(void *client, void (*handler) (const char*, void*), void *handler__context);

/**
 * Event fired when a call is asking for media to send
 * 
 * @param handler takes three arguments, the call Session,
 *        a boolean indicating if audio is desired,
 *        and a boolean indicating if video is desired
 */
API_PREFIX void snikket_client_add_call_media_listener(void *client, void (*handler) (void*, bool, bool, void*), void *handler__context);

/**
 * Event fired when call has a new MediaStreamTrack to play
 * 
 * @param handler takes three arguments, the associated Chat ID,
 *        the new MediaStreamTrack, and an array of any associated MediaStreams
 */
API_PREFIX void snikket_client_add_call_track_listener(void *client, void (*handler) (const char*, void*, void**, size_t, void*), void *handler__context);

/**
 * Let the SDK know the UI is in the foreground
 */
API_PREFIX void snikket_client_set_in_foreground(void *client);

/**
 * Let the SDK know the UI is in the foreground
 */
API_PREFIX void snikket_client_set_not_in_foreground(void *client);

/**
 * ID of this Chat
 */
API_PREFIX const char *snikket_chat_chat_id(void *chat);

/**
 * Current state of this chat
 */
API_PREFIX enum snikket_ui_state snikket_chat_ui_state(void *chat);

/**
 * Fetch a page of messages before some point
 * 
 * @param beforeId id of the message to look before
 * @param beforeTime timestamp of the message to look before,
 *        String in format YYYY-MM-DDThh:mm:ss[.sss]+00:00
 * @param handler takes one argument, an array of ChatMessage that are found
 */
API_PREFIX void snikket_chat_get_messages(void *chat, const char *beforeId, const char *beforeTime, void (*handler) (void**, size_t, void*), void *handler__context);

/**
 * Send a ChatMessage to this Chat
 * 
 * @param message the ChatMessage to send
 */
API_PREFIX void snikket_chat_send_message(void *chat, void *message);

/**
 * Signals that all messages up to and including this one have probably
 * been displayed to the user
 * 
 * @param message the ChatMessage most recently displayed
 */
API_PREFIX void snikket_chat_mark_read_up_to(void *chat, void *message);

/**
 * Save this Chat on the server
 */
API_PREFIX void snikket_chat_bookmark(void *chat);

/**
 * Get the list of IDs of participants in this Chat
 * 
 * @returns array of IDs
 */
API_PREFIX size_t snikket_chat_get_participants(void *chat, const char ***outPtr);

/**
 * Get the details for one participant in this Chat
 * 
 * @param participantId the ID of the participant to look up
 * @param callback takes two arguments, the display name and the photo URI
 */
API_PREFIX void snikket_chat_get_participant_details(void *chat, const char *participantId, void (*callback) (const char*, const char*, void*), void *callback__context);

/**
 * Correct an already-send message by replacing it with a new one
 * 
 * @param localId the localId of the message to correct
 *        must be the localId of the first version ever sent, not a subsequent correction
 * @param message the new ChatMessage to replace it with
 */
API_PREFIX void snikket_chat_correct_message(void *chat, const char *localId, void *message);

/**
 * Add new reaction to a message in this Chat
 * 
 * @param m ChatMessage to react to
 * @param reaction emoji of the reaction
 */
API_PREFIX void snikket_chat_add_reaction(void *chat, void *m, const char *reaction);

/**
 * Remove an already-sent reaction from a message
 * 
 * @param m ChatMessage to remove the reaction from
 * @param reaction the emoji to remove
 */
API_PREFIX void snikket_chat_remove_reaction(void *chat, void *m, const char *reaction);

/**
 * Archive this chat
 */
API_PREFIX void snikket_chat_close(void *chat);

/**
 * An ID of the most recent message in this chat
 */
API_PREFIX const char *snikket_chat_last_message_id(void *chat);

/**
 * The timestamp of the most recent message in this chat
 */
API_PREFIX const char *snikket_chat_last_message_timestamp(void *chat);

/**
 * Get an image to represent this Chat
 * 
 * @param callback takes one argument, the URI to the image
 */
API_PREFIX void snikket_chat_get_photo(void *chat, void (*callback) (const char*, void*), void *callback__context);

/**
 * An ID of the last message displayed to the user
 */
API_PREFIX const char *snikket_chat_read_up_to(void *chat);

/**
 * The number of message that have not yet been displayed to the user
 */
API_PREFIX int snikket_chat_unread_count(void *chat);

/**
 * A preview of the chat, such as the most recent message body
 */
API_PREFIX const char *snikket_chat_preview(void *chat);

/**
 * The display name of this Chat
 */
API_PREFIX const char *snikket_chat_get_display_name(void *chat);

/**
 * Is this a chat with an entity we trust to see our online status?
 */
API_PREFIX bool snikket_chat_is_trusted(void *chat);

API_PREFIX bool snikket_chat_syncing(void *chat);

/**
 * Can audio calls be started in this Chat?
 */
API_PREFIX bool snikket_chat_can_audio_call(void *chat);

/**
 * Can video calls be started in this Chat?
 */
API_PREFIX bool snikket_chat_can_video_call(void *chat);

/**
 * Start a new call in this Chat
 * 
 * @param audio do we want audio in this call
 * @param video do we want video in this call
 */
API_PREFIX void snikket_chat_start_call(void *chat, bool audio, bool video);

/**
 * Accept any incoming calls in this Chat
 */
API_PREFIX void snikket_chat_accept_call(void *chat);

/**
 * Hangup or reject any calls in this chat
 */
API_PREFIX void snikket_chat_hangup(void *chat);

/**
 * The current status of a call in this chat
 */
API_PREFIX const char *snikket_chat_call_status(void *chat);

/**
 * A DTMFSender for a call in this chat, or NULL
 */
API_PREFIX void *snikket_chat_dtmf(void *chat);

/**
 * All video tracks in all active calls in this chat
 */
API_PREFIX size_t snikket_chat_video_tracks(void *chat, void ***outPtr);

/**
 * @returns a new blank ChatMessage
 */
API_PREFIX void *snikket_chat_message_new();

/**
 * The ID as set by the creator of this message
 */
API_PREFIX const char *snikket_chat_message_local_id(void *chat_message);

/**
 * The ID as set by the creator of this message
 */
API_PREFIX void snikket_chat_message_set_local_id(void *chat_message, const char *value);

/**
 * The ID as set by the authoritative server
 */
API_PREFIX const char *snikket_chat_message_server_id(void *chat_message);

/**
 * The ID as set by the authoritative server
 */
API_PREFIX void snikket_chat_message_set_server_id(void *chat_message, const char *value);

/**
 * The ID of the server which set the serverId
 */
API_PREFIX const char *snikket_chat_message_server_id_by(void *chat_message);

/**
 * The ID of the server which set the serverId
 */
API_PREFIX void snikket_chat_message_set_server_id_by(void *chat_message, const char *value);

/**
 * The timestamp of this message, in format YYYY-MM-DDThh:mm:ss[.sss]+00:00
 */
API_PREFIX const char *snikket_chat_message_timestamp(void *chat_message);

/**
 * The timestamp of this message, in format YYYY-MM-DDThh:mm:ss[.sss]+00:00
 */
API_PREFIX void snikket_chat_message_set_timestamp(void *chat_message, const char *value);

/**
 * Message this one is in reply to, or NULL
 */
API_PREFIX void *snikket_chat_message_reply_to_message(void *chat_message);

/**
 * Message this one is in reply to, or NULL
 */
API_PREFIX void snikket_chat_message_set_reply_to_message(void *chat_message, void *value);

/**
 * ID of the thread this message is in, or NULL
 */
API_PREFIX const char *snikket_chat_message_thread_id(void *chat_message);

/**
 * ID of the thread this message is in, or NULL
 */
API_PREFIX void snikket_chat_message_set_thread_id(void *chat_message, const char *value);

/**
 * Array of attachments to this message
 */
API_PREFIX size_t snikket_chat_message_attachments(void *chat_message, void ***outPtr);

/**
 * Body text of this message or NULL
 */
API_PREFIX const char *snikket_chat_message_text(void *chat_message);

/**
 * Body text of this message or NULL
 */
API_PREFIX void snikket_chat_message_set_text(void *chat_message, const char *value);

/**
 * Language code for the body text
 */
API_PREFIX const char *snikket_chat_message_lang(void *chat_message);

/**
 * Language code for the body text
 */
API_PREFIX void snikket_chat_message_set_lang(void *chat_message, const char *value);

/**
 * Is this a Group Chat message?
 * 
 * If the message is in the context of a Channel but this is false,
 * then it is a private message
 */
API_PREFIX bool snikket_chat_message_is_groupchat(void *chat_message);

/**
 * Is this a Group Chat message?
 * 
 * If the message is in the context of a Channel but this is false,
 * then it is a private message
 */
API_PREFIX void snikket_chat_message_set_is_groupchat(void *chat_message, bool value);

/**
 * Direction of this message
 */
API_PREFIX enum snikket_message_direction snikket_chat_message_direction(void *chat_message);

/**
 * Direction of this message
 */
API_PREFIX void snikket_chat_message_set_direction(void *chat_message, enum snikket_message_direction value);

/**
 * Status of this message
 */
API_PREFIX enum snikket_message_status snikket_chat_message_status(void *chat_message);

/**
 * Status of this message
 */
API_PREFIX void snikket_chat_message_set_status(void *chat_message, enum snikket_message_status value);

/**
 * Array of past versions of this message, if it has been edited
 */
API_PREFIX size_t snikket_chat_message_versions(void *chat_message, void ***outPtr);

API_PREFIX void snikket_chat_message_add_attachment(void *chat_message, void *attachment);

/**
 * Create a new ChatMessage in reply to this one
 */
API_PREFIX void *snikket_chat_message_reply(void *chat_message);

/**
 * Get HTML version of the message body
 */
API_PREFIX const char *snikket_chat_message_html(void *chat_message);

/**
 * The ID of the Chat this message is associated with
 */
API_PREFIX const char *snikket_chat_message_chat_id(void *chat_message);

/**
 * The ID of the sender of this message
 */
API_PREFIX const char *snikket_chat_message_sender_id(void *chat_message);

/**
 * The ID of the account associated with this message
 */
API_PREFIX const char *snikket_chat_message_account(void *chat_message);

/**
 * Is this an incoming message?
 */
API_PREFIX bool snikket_chat_message_is_incoming(void *chat_message);

/**
 * The URI of an icon for the thread associated with this message, or NULL
 */
API_PREFIX const char *snikket_chat_message_thread_icon(void *chat_message);

/**
 * Duplicate this ChatMessage
 */
API_PREFIX void *snikket_chat_message_clone(void *chat_message);

API_PREFIX const char *snikket_chat_attachment_name(void *chat_attachment);

API_PREFIX const char *snikket_chat_attachment_mime(void *chat_attachment);

API_PREFIX int snikket_chat_attachment_size(void *chat_attachment);

API_PREFIX size_t snikket_chat_attachment_uris(void *chat_attachment, const char ***outPtr);

API_PREFIX const char *snikket_jingle_media_stream_track_id(void *media_stream_track);

API_PREFIX bool snikket_jingle_media_stream_track_muted(void *media_stream_track);

API_PREFIX const char *snikket_jingle_media_stream_track_kind(void *media_stream_track);

API_PREFIX size_t snikket_jingle_media_stream_track_supported_audio_formats(void *media_stream_track, void ***outPtr);

/**
 * Event fired for new inbound audio frame
 * 
 * @param callback takes three arguments, the Signed 16-bit PCM data, the clock rate, and the number of channels
 */
API_PREFIX void snikket_jingle_media_stream_track_add_pcm_listener(void *media_stream_track, void (*callback) (short*, size_t, int, int, void*), void *callback__context);

/**
 * Event fired when ready for next outbound audio frame
 * 
 * @param callback
 */
API_PREFIX void snikket_jingle_media_stream_track_add_ready_for_pcm_listener(void *media_stream_track, void (*callback) (void*), void *callback__context);

/**
 * Send new audio to this track
 * 
 * @param pcm 16-bit signed linear PCM data (interleaved)
 * @param clockRate the sampling rate of the data
 * @param channels the number of audio channels
 */
API_PREFIX void snikket_jingle_media_stream_track_write_pcm(void *media_stream_track, const short *pcm, size_t pcm__len, int clockRate, int channels);

API_PREFIX void snikket_jingle_media_stream_track_stop(void *media_stream_track);

API_PREFIX void *snikket_jingle_audio_format_new(const char *format, unsigned char payloadType, int clockRate, int channels);

API_PREFIX int snikket_jingle_audio_format_clock_rate(void *audio_format);

API_PREFIX int snikket_jingle_audio_format_channels(void *audio_format);

/**
 * The ID of the Chat this search result represents
 */
API_PREFIX const char *snikket_available_chat_chat_id(void *available_chat);

/**
 * The display name of this search result
 */
API_PREFIX const char *snikket_available_chat_display_name(void *available_chat);

/**
 * A human-readable note associated with this search result
 */
API_PREFIX const char *snikket_available_chat_note(void *available_chat);

/**
 * Is this search result a channel?
 */
API_PREFIX bool snikket_available_chat_is_channel(void *available_chat);

API_PREFIX const char *snikket_jingle_initiated_session_sid(void *initiated_session);

API_PREFIX void snikket_jingle_initiated_session_accept(void *initiated_session);

API_PREFIX void snikket_jingle_initiated_session_hangup(void *initiated_session);

API_PREFIX void snikket_jingle_initiated_session_add_media(void *initiated_session, void *const *streams, size_t streams__len);

API_PREFIX const char *snikket_jingle_initiated_session_call_status(void *initiated_session);

API_PREFIX size_t snikket_jingle_initiated_session_video_tracks(void *initiated_session, void ***outPtr);

API_PREFIX void *snikket_jingle_initiated_session_dtmf(void *initiated_session);

API_PREFIX void snikket_jingle_initiated_session_supply_media(void *initiated_session, void *const *streams, size_t streams__len);

API_PREFIX void *snikket_jingle_media_stream_new();

/**
 * Create default bidirectional audio track
 */
API_PREFIX void *snikket_jingle_media_stream_make_audio();

API_PREFIX void snikket_jingle_media_stream_add_track(void *media_stream, void *track);

API_PREFIX size_t snikket_jingle_media_stream_get_tracks(void *media_stream, void ***outPtr);

/**
 * Schedule DTMF events to be sent
 * 
 * @param tones can be any number of 0123456789#*ABCD,
 */
API_PREFIX void snikket_jingle_dtmf_sender_insert_dtmf(void *dtmf_sender, const char *tones);

#ifdef __cplusplus
}
#endif

#undef API_PREFIX

#endif
