const std = @import("std");
const CallStatus = @import("libhoneybee").state.CallStatus;

const s_api = @import("snikket_api.zig"); // transpiled using  zig translate-c -I /usr/include/  > src/snikket_api.zig
const a_be_c = @cImport(@cInclude("audio_be.c"));

const StateChange = @import("state.zig").StateChange;
const CallbackFifo = @import("state.zig").CallbackFifo;

const SnikketCallStatus = enum { none, outgoing };

fn panic(info: [*c]const u8) callconv(.C) void {
    std.debug.panic("SDK panic {s}\n", .{info});
}

fn retract_call(jid: [*c]const u8, cont: ?*anyopaque) callconv(.C) void {
    const context: *CallbackFifo = @ptrCast(@alignCast(cont));
    const new_state_change = StateChange{ .new_state = .end_call, .jid = jid };
    context.*.add_message(new_state_change);
}

fn offline(cont: ?*anyopaque) callconv(.C) void {
    const context: *CallbackFifo = @ptrCast(@alignCast(cont));
    const new_state_change = StateChange{
        .new_state = .offline,
    };
    context.*.add_message(new_state_change);
}

fn on_password_needed(client: ?*anyopaque, password: ?*anyopaque) callconv(.C) void {
    s_api.snikket_client_use_password(client, @ptrCast(password));
    s_api.snikket_release(client);
}

fn on_ready(cont: ?*anyopaque) callconv(.C) void {
    const context: *CallbackFifo = @ptrCast(@alignCast(cont));
    const new_state_change = StateChange{
        .new_state = .online,
    };
    context.*.add_message(new_state_change);
}

fn on_call_ringing(jid: [*c]const u8, cont: ?*anyopaque) callconv(.C) void {
    // fired when outgoing call is ringing
    const context: *CallbackFifo = @ptrCast(@alignCast(cont));
    const new_state_change = StateChange{ .new_state = .start_call, .jid = jid };
    context.*.add_message(new_state_change);
}

fn on_call_ring(session: ?*anyopaque, jid: [*c]const u8, cont: ?*anyopaque) callconv(.C) void {
    const context: *CallbackFifo = @ptrCast(@alignCast(cont));
    defer s_api.snikket_release(session);

    std.log.info("Incoming call {s}", .{jid});
    const new_state_change = StateChange{ .new_state = .incoming_call, .jid = jid };
    context.*.add_message(new_state_change);
}

fn on_ready_for_pcm(dummy: ?*anyopaque) callconv(.C) void {
    a_be_c.on_ready_for_pcm(dummy);
}

fn on_pcm(pcm: [*c]c_short, cpcm: usize, clock_rate: c_int, channels: c_int, pa_in: ?*anyopaque) callconv(.C) void {
    const clock_rate_size: c_int = @intCast(clock_rate);
    const channels_size: u8 = @intCast(channels);
    // const pa: ?*a_be_c.struct_pa_simple = @ptrCast(pa_in);
    _ = pa_in;
    a_be_c.on_pcm(pcm, cpcm, clock_rate_size, channels_size);
}

fn on_call_media(session: ?*anyopaque, audio: bool, video: bool, cont: ?*anyopaque) callconv(.C) void {
    defer s_api.snikket_release(session);
    var streams = [_]?*anyopaque{ null, null };
    const pa_in: ?*anyopaque = @ptrCast(@alignCast(cont));
    _ = audio;
    _ = video;

    var tracks: [*c]?*anyopaque = null;
    streams[0] = s_api.snikket_jingle_media_stream_make_audio();
    const ctracks = s_api.snikket_jingle_media_stream_get_tracks(streams[0], @ptrCast(&tracks));
    const cstreams = 1;
    for (0..ctracks) |i| {
        s_api.snikket_jingle_media_stream_track_add_pcm_listener(tracks[i], on_pcm, pa_in);
        s_api.snikket_jingle_media_stream_track_add_ready_for_pcm_listener(tracks[i], on_ready_for_pcm, tracks[i]);
        s_api.snikket_release(tracks[i]);
    }
    s_api.snikket_jingle_initiated_session_supply_media(session, &streams[0], cstreams);

    for (0..ctracks) |i| {
        s_api.snikket_release(streams[i]);
    }
}

const ChatContext = struct {
    client: ?*?*anyopaque,
    chat: ?*?*anyopaque,
};

fn available_chats(q: [*c]const u8, chats: [*c]?*anyopaque, cchats: usize, cont: ?*anyopaque) callconv(.C) void {
    const context: ?*ChatContext = @ptrCast(@alignCast(cont));
    const client = context.?.*.client.?.*;
    switch (cchats) {
        0 => std.log.err("No call target found query: {s}", .{q}),
        1 => blk: {
            context.?.chat.?.* = s_api.snikket_client_start_chat(client, chats[0]);
            break :blk;
        },
        else => std.log.err("Unknown call target query: {s}", .{q}), // TODO: why does this keep running?
    }

    for (0..cchats) |i| {
        s_api.snikket_release(chats[i]);
    }
    s_api.snikket_release(@ptrCast(chats));
    s_api.snikket_release(q);
}

pub const CallState = struct {
    is_outbound_only: bool = false, // outbound (dont receive calls)
    pa_backend: ?*anyopaque = null,

    messages: *CallbackFifo,
    status: CallStatus = CallStatus.uninitialized,
    client: ?*anyopaque = null,
    incall_jid: [*c]const u8 = null,

    pub fn init(self: *CallState, username: [:0]const u8, password: [:0]const u8) void {
        // need null terminated strings for username and password because we are passing both to c functions
        std.debug.assert(self.status == CallStatus.uninitialized);

        _ = s_api.snikket_setup(panic);

        const persistence = s_api.snikket_persistence_dummy_new();
        const client = s_api.snikket_client_new(@ptrCast(username), persistence);
        self.client = client;
        s_api.snikket_client_set_send_available(self.client, !self.is_outbound_only);
        s_api.snikket_client_add_password_needed_listener(self.client, on_password_needed, @ptrCast(@constCast(password)));
        s_api.snikket_client_add_call_ringing_listener(self.client, on_call_ringing, self.messages);
        s_api.snikket_client_add_call_ring_listener(self.client, on_call_ring, self.messages);

        s_api.snikket_client_add_call_retract_listener(self.client, retract_call, self.messages);
        s_api.snikket_client_add_status_offline_listener(self.client, offline, self.messages);
        s_api.snikket_client_add_status_online_listener(self.client, on_ready, self.messages);
        s_api.snikket_client_add_call_media_listener(self.client, on_call_media, self.pa_backend);
        self.messages.add_message(StateChange{ .new_state = .offline });
        std.debug.assert(self.messages.len() == 1);
    }

    pub fn connect(self: *CallState) void {
        std.debug.assert(self.status == CallStatus.unconnect);

        std.log.info("connecting", .{});
        self.messages.clear();
        s_api.snikket_client_start(self.client);
        while (self.messages.len() == 0) {
            std.time.sleep(std.time.ns_per_s); // Why doesnt api let me wait till it's connected?
        }
    }

    fn get_chat_id(self: *CallState, call_jid: [:0]const u8) ?*anyopaque {
        var chat_id: ?*anyopaque = null;
        const chat_context = ChatContext{ .client = &self.client, .chat = &chat_id };
        s_api.snikket_client_find_available_chats(self.client, call_jid, available_chats, @ptrCast(@constCast(&chat_context)));
        std.time.sleep(std.time.ns_per_s * 3);
        return chat_id;
    }

    pub fn call(self: *CallState, call_jid: [:0]const u8) void {
        std.debug.assert(self.status == CallStatus.connected);
        // TODO check if  incoming call and just accept call
        // const char *snikket_chat_call_status(void *chat);
        // void snikket_chat_accept_call(void *chat);

        std.log.info("call target: {s}", .{call_jid});

        const chat_id = self.get_chat_id(call_jid);

        if (chat_id == null) {
            // TODO raise error
            std.log.err("Could not find {s}", .{call_jid});
            return;
        }

        s_api.snikket_chat_start_call(chat_id, true, false);
    }

    pub fn accept_call(self: *CallState, call_jid: [:0]const u8) void {
        std.debug.assert(self.status == CallStatus.connected);
        // const call_val: [*c]const u8 = @ptrCast(call_jid);
        // const chat_id = s_api.snikket_client_get_chat(self.client, call_val); TODO: unless call_val = literal string, this function returns null
        const chat_id = self.get_chat_id(call_jid);
        std.debug.assert(chat_id != null);

        s_api.snikket_chat_accept_call(chat_id);
    }

    pub fn hangup(self: *CallState, call_jid: [:0]const u8) void {
        std.debug.assert(self.status == CallStatus.incall);

        std.log.info("hanging up", .{});
        const chat_id = self.get_chat_id(call_jid);
        std.debug.assert(chat_id != null);
        s_api.snikket_chat_hangup(chat_id);

        std.Thread.sleep(0.5 * std.time.ns_per_s); // make sure snikket api has hung up before I check if retracted
        // need to call retract call myself
        const status = std.meta.stringToEnum(SnikketCallStatus, std.mem.span(s_api.snikket_chat_call_status(chat_id)));
        if (status == SnikketCallStatus.none) {
            retract_call(call_jid, self.messages);
        }
    }

    pub fn send_dtmf(self: *CallState, tones: [:0]const u8) void {
        const tone: *const [1:0]u8 = &tones[0];
        std.debug.assert(self.status == CallStatus.incall);
        std.debug.assert(self.incall_jid != null);
        const chat_id = self.get_chat_id(std.mem.sliceTo(self.incall_jid, 0));
        const sender = s_api.snikket_chat_dtmf(chat_id);
        s_api.snikket_jingle_dtmf_sender_insert_dtmf(sender, tone);
    }
};
