const std = @import("std");

const EventType = @import("libhoneybee").state.EventType;
const SocketFifo = @import("state.zig").SocketFifo;

pub const Event = struct {
    event_type: EventType,
    payload: [:0]const u8,
    allocator: std.mem.Allocator,

    pub fn free(self: *Event) void {
        self.allocator.free(self.payload);
    }
};

const socket_name = "honeybee.sock";
const MAX_JID_LENGTH = 128;

pub const SockHost = struct {
    address: std.net.Address,
    sock: i32,
    allocator: std.mem.Allocator,
    messages: *SocketFifo,

    pub fn connect(self: *SockHost) !void {
        try std.posix.bind(self.sock, &self.address.any, self.address.getOsSockLen());
    }

    pub fn write_event(self: *SockHost, event: EventType, jid: [:0]const u8) !void {
        const buf_size = 1024;
        var buf: [buf_size]u8 = undefined; // TODO: make this buffer smaller
        const command = try std.fmt.bufPrint(&buf, "{s};{s}", .{ @tagName(event), jid });

        std.log.debug(
            "sending: {s}",
            .{command},
        );

        _ = try std.posix.send(
            self.sock,
            command,
            0,
        );
    }

    pub fn is_daemon_active(self: *SockHost) bool {
        std.posix.bind(self.sock, &self.address.any, self.address.getOsSockLen()) catch |err| switch (err) {
            error.AddressInUse => return true,
            error.AlreadyBound => return true,
            else => unreachable,
        };
        return false;
    }

    pub fn block_till_event(self: *SockHost) !void {
        var other_addr: std.posix.sockaddr = undefined;
        var other_addrlen: std.posix.socklen_t = @sizeOf(std.posix.sockaddr);
        var buf: [1024]u8 = undefined;

        // we did not set the NONBLOCK flag (socket type flag),
        // so the program will wait until data is received
        const n_recv = try std.posix.recvfrom(
            self.sock,
            buf[0..],
            0,
            &other_addr,
            &other_addrlen,
        );

        const dat = buf[0..n_recv];
        var s_dat = std.mem.splitScalar(u8, dat, ';');

        std.log.debug(
            "received {d} byte(s) from {any};\n    string: {s}\n",
            .{ n_recv, other_addr, buf[0..n_recv] },
        );

        const id: []const u8 = s_dat.first();
        var event_type = EventType.invalid;
        if (std.mem.eql(u8, "call", id)) {
            event_type = EventType.call;
        } else if (std.mem.eql(u8, "hangup", id)) {
            event_type = EventType.hangup;
        } else if (std.mem.eql(u8, "dtmf", id)) {
            event_type = EventType.dtmf;
        } else if (std.mem.eql(u8, "accept_call", id)) {
            event_type = EventType.accept_call;
        }
        const payload_to_store = s_dat.peek().?;

        const payload = try self.allocator.allocSentinel(u8, payload_to_store.len, 0);
        @memcpy(payload, s_dat.peek().?);

        const event = Event{
            .event_type = event_type,
            .payload = payload,
            .allocator = self.allocator,
        };
        self.messages.*.add_message(event);
    }
};
