const std = @import("std");
const s_api = @import("snikket_api.zig");
const SocketService = @import("socket_service.zig");
const s_call = @import("snikket_call.zig");
const state_common = @import("libhoneybee").state;

pub const HookHandler = struct {
    hook_path: []const u8,
    allocator: std.mem.Allocator,

    pub fn run_hook(self: *HookHandler, hook: state_common.HookType, jid: [*c]const u8) void {
        const hook_full_path: []const u8 = std.fs.path.join(self.allocator, &[_][]const u8{ self.hook_path, "honeybee", @tagName(hook) }) catch unreachable;
        defer self.allocator.free(hook_full_path);

        const jid_z: []u8 = std.mem.span(@constCast(jid));
        var cmd = std.process.Child.init(&[_][]const u8{ hook_full_path, jid_z }, self.allocator);
        cmd.spawn() catch unreachable;
        std.log.debug("Running hook {s}", .{hook_full_path});
        // We may not want to fire and forget here and instead log the output of script
        // However to make this asynchronous, we would have to do this in a seperate thread
        // _ = cmd.wait() catch unreachable;
    }
};

pub const StateChange = struct {
    new_state: state_common.HookType,
    // Set these to default values, it is up to the function that applies this state change to ignore values
    // that dont matter
    jid: [*c]const u8 = null,

    pub fn free(self: *StateChange) void {
        if (self.jid != null) {
            s_api.snikket_release(self.jid);
        }
    }
};

fn MessageFifo(
    comptime T: type,
) type {
    return struct {
        q: std.fifo.LinearFifo(T, .Dynamic),
        mutex: *std.Thread.Mutex,
        cond: *std.Thread.Condition,

        const Self = @This();

        pub fn clear(self: *Self) void {
            while (self.len() > 0) {
                _ = self.next_message();
            }
        }

        pub fn add_message(self: *Self, msg: T) void {
            self.mutex.lock();
            self.q.writeItem(msg) catch unreachable;
            self.mutex.unlock();
            std.log.debug("sending signal", .{});
            self.cond.signal();
        }

        pub fn next_message(self: *Self) T {
            self.mutex.lock();
            defer self.mutex.unlock();
            return self.q.readItem().?;
        }

        pub fn len(self: *Self) usize {
            return self.q.readableLength();
        }

        pub fn deinit(self: *Self) void {
            self.mutex.lock();
            defer self.mutex.unlock();
            return self.q.deinit();
        }
    };
}

pub const CallbackFifo = MessageFifo(StateChange);
pub const SocketFifo = MessageFifo(SocketService.Event);

pub fn service_message(callback_fifo: *CallbackFifo, hook_handler: *HookHandler, state_file: *state_common.StateFile, call_state: *s_call.CallState) !void {
    var message = callback_fifo.next_message();
    defer message.free();

    // change state of snikket callstate
    switch (message.new_state) {
        state_common.HookType.online,
        => {
            call_state.status = state_common.CallStatus.connected;
            call_state.incall_jid = null;
        },
        state_common.HookType.offline => {
            call_state.status = state_common.CallStatus.unconnect;
            call_state.incall_jid = null;
        },
        state_common.HookType.start_call => {
            call_state.status = state_common.CallStatus.incall;
            call_state.incall_jid = message.jid;
        },
        state_common.HookType.incoming_call => {},
        state_common.HookType.end_call => {
            call_state.status = state_common.CallStatus.connected;
            call_state.incall_jid = null;
        },
    }

    if (message.jid != null) {
        hook_handler.run_hook(message.new_state, message.jid);
    } else {
        hook_handler.run_hook(message.new_state, "");
    }
    try state_file.update_state_file(call_state.status, message.jid);
}
