const std = @import("std");
const state = @import("libhoneybee").state;
const config = @import("libhoneybee").config;
const socket = @import("libhoneybee").socket;

const SockClient = struct {
    address: std.net.Address,
    sock: i32,

    pub fn is_daemon_active(self: *SockClient) bool {
        std.posix.bind(self.sock, &self.address.any, self.address.getOsSockLen()) catch |err| switch (err) {
            error.AddressInUse => return true,
            error.AlreadyBound => return true,
            else => unreachable,
        };
        return false;
    }

    pub fn connect(self: *SockClient) !void {
        try std.posix.connect(self.sock, &self.address.any, self.address.getOsSockLen());
    }

    pub fn write_event(self: *SockClient, event: state.EventType, jid: [:0]const u8) !void {
        const buf_size = 1024;
        var buf: [buf_size]u8 = undefined; // TODO: make this buffer smaller
        const command = try std.fmt.bufPrint(&buf, "{s};{s}", .{ @tagName(event), jid });

        std.log.debug(
            "sending: {s}",
            .{command},
        );

        _ = try std.posix.send(
            self.sock,
            command,
            0,
        );
    }
};

pub fn print_help() void {
    std.log.info("usage:", .{});
    std.log.info("status: print status of daemon", .{});
    std.log.info("jid: print jid you are currently calling", .{});
    std.log.info("call <jid>: call <jid>", .{});
    std.log.info("hangup <jid>: hangup from <jid>", .{});
    std.log.info("accept_call <jid>: accept call from <jid>", .{});
    std.log.info("dtmf <dtmf_tone>: send <dtmf_tone> to active call", .{});
}

pub fn main() !void {
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    defer _ = gpa.deinit();
    const allocator = gpa.allocator();

    const conf = try config.get_env_variables(allocator);
    defer conf.free();

    const sock_info = socket.get_sock_info();

    var sock = SockClient{ .address = sock_info.address, .sock = sock_info.sock };
    const is_daemon_active = sock.is_daemon_active();
    if (!is_daemon_active) {
        std.log.err("Honeybee daemon is not running", .{});
        std.process.exit(1);
    }

    const state_file = state.StateFile{
        .runtime_dir = conf.xdg_runtime_dir,
        .state_path = conf.state_path,
        .allocator = allocator,
    };
    const cur_status = try state_file.get_state();
    defer cur_status.free();
    try sock.connect();

    var args_no_alloc_iter = std.process.args();
    _ = args_no_alloc_iter.next();
    const command = args_no_alloc_iter.next() orelse "";

    if (std.mem.eql(u8, "status", command)) {
        std.debug.print("{}\n", .{cur_status.status});
    } else if (std.mem.eql(u8, "jid", command)) {
        if (cur_status.jid) |jid| {
            std.debug.print("{s}\n", .{jid});
        } else {
            std.log.err("No jid in current status {}", .{cur_status.status});
            std.process.exit(1);
        }
    } else if (std.mem.eql(u8, "call", command)) {
        const send_cmd = state.EventType.call;
        if (!state.is_valid(cur_status.status, send_cmd)) {
            std.log.err("Current Status: {}. Cannot Call right now", .{cur_status});
            std.process.exit(1);
        }
        const jid = args_no_alloc_iter.next();
        if (jid == null) {
            std.log.err("You did not specify jid", .{});
            std.process.exit(1);
        }
        try sock.write_event(send_cmd, jid.?);
    } else if (std.mem.eql(u8, "hangup", command)) {
        const send_cmd = state.EventType.hangup;
        if (!state.is_valid(cur_status.status, send_cmd)) {
            std.log.err("Current Status: {}. Cannot hangup right now", .{cur_status});
            std.process.exit(1);
        }
        const jid = args_no_alloc_iter.next();
        if (jid == null) {
            std.log.err("You did not specify jid", .{});
            std.process.exit(1);
        }
        try sock.write_event(send_cmd, jid.?);
    } else if (std.mem.eql(u8, "accept_call", command)) {
        const send_cmd = state.EventType.accept_call;
        if (!state.is_valid(cur_status.status, send_cmd)) {
            std.log.err("Current Status: {}. Cannot accept_call right now", .{cur_status});
            std.process.exit(1);
        }
        const jid = args_no_alloc_iter.next();
        if (jid == null) {
            std.log.err("You did not specify jid", .{});
            std.process.exit(1);
        }
        try sock.write_event(send_cmd, jid.?);
    } else if (std.mem.eql(u8, "dtmf", command)) {
        const send_cmd = state.EventType.dtmf;
        if (!state.is_valid(cur_status.status, send_cmd)) {
            std.log.err("Current Status: {}. Cannot send dtmf right now", .{cur_status});
            std.process.exit(1);
        }
        const dtmf_tone = args_no_alloc_iter.next();
        if (dtmf_tone == null) {
            std.log.err("You did not specify dtmf_tone", .{});
            std.process.exit(1);
        }
        try sock.write_event(send_cmd, dtmf_tone.?);
    } else {
        print_help();
    }
}
