const std = @import("std");

// Events that the user can send to daemon
pub const EventType = enum { call, hangup, dtmf, accept_call, invalid };
// States the call daemon can be in
pub const CallStatus = enum { uninitialized, unconnect, connected, incall };
// Hooks - Events that users can bind to
pub const HookType = enum { start_call, incoming_call, end_call, online, offline };

pub fn is_valid(cur: CallStatus, event: EventType) bool {
    if (cur == CallStatus.uninitialized or cur == CallStatus.unconnect) {
        return false;
    }

    if (cur == CallStatus.connected and (event == EventType.hangup or event == EventType.dtmf or event == EventType.invalid)) {
        return false;
    }

    if (cur == CallStatus.incall and (event == EventType.accept_call or event == EventType.call or event == EventType.invalid)) {
        return false;
    }

    return true;
}

const StateJid = struct {
    full_str: []u8,
    status: CallStatus,
    jid: ?[]const u8,
    allocator: std.mem.Allocator,

    pub fn free(self: *const StateJid) void {
        self.allocator.free(self.full_str);
    }
};

pub const StateFile = struct {
    state_path: []const u8,
    runtime_dir: []const u8,
    allocator: std.mem.Allocator,

    pub fn create_path(self: *const StateFile) void {
        // create path required for state file
        // Should only be run once
        const hb_runtime_path: []const u8 = std.fs.path.join(self.allocator, &[_][]const u8{ self.runtime_dir, "honeybee" }) catch unreachable;
        defer self.allocator.free(hb_runtime_path);

        std.log.debug("Making dir {s}", .{hb_runtime_path});
        std.fs.makeDirAbsolute(hb_runtime_path) catch |err| switch (err) {
            std.posix.MakeDirError.PathAlreadyExists => return,
            else => unreachable,
        };
    }

    pub fn update_state_file(self: *const StateFile, status: CallStatus, jid: [*c]const u8) !void {
        const buf_size = 1024;
        var buf: [buf_size]u8 = undefined; // TODO: make this buffer smaller

        var status_jid: []u8 ="";
        if (jid == null) {
            status_jid = try std.fmt.bufPrint(&buf, "{s}", .{ @tagName(status)});
        } else {
            status_jid = try std.fmt.bufPrint(&buf, "{s};{s}", .{ @tagName(status), jid });
        }

        std.log.debug("Updating {s} to: {s}", .{ self.state_path, status_jid });
        const file = try std.fs.createFileAbsolute(self.state_path, .{ .truncate = true, .exclusive = false });
        defer file.close();

        _ = file.writeAll(status_jid) catch unreachable;
        std.log.debug("Updated {s} to: {s}", .{ self.state_path, status_jid });
    }

    pub fn get_state(self: *const StateFile) !StateJid {
        const file = try std.fs.openFileAbsolute(self.state_path, .{});
        defer file.close();

        const buffer_size = 2000; // TODO: can probably lower
        const file_buffer = try file.readToEndAlloc(self.allocator, buffer_size);

        var s_dat = std.mem.splitScalar(u8, file_buffer, ';');
        const status_str: []const u8 = s_dat.first();
        const jid_str: ?[]const u8 = s_dat.next();

        const status = std.meta.stringToEnum(CallStatus, status_str).?;

        return StateJid{
            .full_str = file_buffer,
            .allocator = self.allocator,
            .status = status,
            .jid = jid_str,
        };
    }
};
